--[[
Name: DogTag-1.0
Revision: $Rev: 32379 $
Author: Cameron Kenneth Knight (ckknight@gmail.com)
Website: http://wiki.wowace.com/index.php/DogTag-1.0
Documentation: http://wiki.wowace.com/index.php/DogTag-1.0
SVN: svn://svn.wowace.com/wowace/trunk/DogTag-1.0/DogTag-1.0
Description: A library to provide a markup syntax for unit frame FontStrings
Dependencies: AceLibrary, AceOO-2.0, AceEvent-2.0
License: LGPL v2.1
]]

local MAJOR_VERSION = "DogTag-1.0"
local MINOR_VERSION = "$Revision: 32379 $"

-- This ensures the code is only executed if the libary doesn't already exist, or is a newer version
if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

if not AceLibrary:HasInstance("AceOO-2.0") then error(MAJOR_VERSION .. " requires AceOO-2.0.") end
if not AceLibrary:HasInstance("AceEvent-2.0") then error(MAJOR_VERSION .. " requires AceEvent-2.0.") end

local AceEvent = AceLibrary("AceEvent-2.0")

local DogTag = {}

local L = {
	-- short races
	["BE"] = "BE",
	["Dr"] = "Dr",
	["Dw"] = "Dw",
	["Gn"] = "Gn",
	["Hu"] = "Hu",
	["NE"] = "NE",
	["Or"] = "Or",
	["Ta"] = "Ta",
	["Tr"] = "Tr",
	["Ud"] = "Ud",
	
	-- short classes
	["Wr"] = "Wr",
	["Pr"] = "Pr",
	["Ma"] = "Ma",
	["Sh"] = "Sh",
	["Pa"] = "Pa",
	["Wl"] = "Wl",
	["Dr"] = "Dr",
	["Ro"] = "Ro",
	["Hu"] = "Hu",
	
	-- classifications
	["Rare"] = ITEM_QUALITY3_DESC,
	["Rare-Elite"] = ITEM_QUALITY3_DESC .. "-" .. ELITE,
	["Elite"] = ELITE,
	["Boss"] = BOSS,
	-- short classifications
	["r"] = "r",
	["r+"] = "r+",
	["+"] = "+",
	["b"] = "b",
	
	["Dead"] = DEAD,
	["Ghost"] = "Ghost",
	["Offline"] = PLAYER_OFFLINE,
	["Combat"] = "Combat",
	["Resting"] = "Resting",
	["Tapped"] = "Tapped",
	
	["Rage"] = RAGE,
	["Focus"] = FOCUS,
	["Energy"] = ENERGY,
	["Mana"] = MANA,
	
	["PvP"] = PVP,
	
	-- genders
	["Male"] = MALE,
	["Female"] = FEMALE,
	
	-- shortgenders
	["m"] = "m",
	["f"] = "f",
	
	["Rested"] = "Rested",
	
	["Leader"] = "Leader",
	["L"] = "L", -- short leader
}
if GetLocale() == "deDE" then
	-- alter L here
end

local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local UNKNOWN = UNKNOWN or "Unknown"
local unpack = unpack

local new, del
do
	local list = setmetatable({}, {__mode='k'})
	function new()
		local t = next(list)
		if t then
			list[t] = nil
			return t
		else
			return {}
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end

local colors = {}
for class, data in pairs(RAID_CLASS_COLORS) do
	colors[class] = { data.r, data.g, data.b, }
end
colors.unknown = { 0.8, 0.8, 0.8 }
colors.hostile = { 226/255, 45/255, 75/255 }
colors.neutral = { 1, 1, 34/255 }
colors.friendly = { 0.2, 0.8, 0.15 }
colors.civilian = { 48/255, 113/255, 191/255 }
colors.tapped = { 0.5, 0.5, 0.5 }
colors.dead = { 0.6, 0.6, 0.6 }
colors.disconnected = { 0.7, 0.7, 0.7 }
colors.inCombat = { 1, 0, 0 }
colors.resting = { 1, 1, 0 }
colors.petHappy = { 0, 1, 0 }
colors.petNeutral = { 1, 1, 0 }
colors.petAngry = { 1, 0, 0 }
colors.rage = { 226/255, 45/255, 75/255 }
colors.energy = { 1, 220/255, 25/255 }
colors.focus = { 1, 210/255, 0 }
colors.mana = { 48/255, 113/255, 191/255 }

local function GetClassColor(unit)
	local _,class = UnitClass(unit)
	return unpack(colors[class] or colors.unknown)
end

local function GetHostilityColor(unit)
	if UnitIsPlayer(unit) or UnitPlayerControlled(unit) then
		if UnitCanAttack(unit, "player") then
			-- they can attack me
			if UnitCanAttack("player", unit) then
				-- and I can attack them
				return unpack(colors.hostile)
			else
				-- but I can't attack them
				return unpack(colors.civilian)
			end
		elseif UnitCanAttack("player", unit) then
			-- they can't attack me, but I can attack them
			return unpack(colors.neutral)
		elseif UnitIsPVP(unit) then
			-- on my team
			return unpack(colors.friendly)
		else
			-- either enemy or friend, no violence
			return unpack(colors.civilian)
		end
	elseif (UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit)) or UnitIsDead(unit) then
		return unpack(colors.tapped)
	else
		local reaction = UnitReaction(unit, "player")
		if reaction then
			if reaction >= 5 then
				return unpack(colors.friendly)
			elseif reaction == 4 then
				return unpack(colors.neutral)
			else
				return unpack(colors.hostile)
			end
		else
			return unpack(colors.unknown)
		end
	end
end

-- Note, these tables use the unlocalized second returns of UnitClass and UnitRace, thus the keys are not localized.
local shortraces = {
	BloodElf = L["BE"],
	Draenei = L["Dr"],
	Dwarf = L["Dw"],
	Gnome = L["Gn"],
	Human = L["Hu"],
	NightElf = L["NE"],
	Orc = L["Or"],
	Tauren = L["Ta"],
	Troll = L["Tr"],
	Undead = L["Ud"],
}
local shortclasses = {
	WARRIOR = L["Wr"],
	PRIEST = L["Pr"],
	MAGE = L["Ma"],
	SHAMAN = L["Sh"],
	PALADIN = L["Pa"],
	WARLOCK = L["Wl"],
	DRUID = L["Dr"],
	ROGUE = L["Ro"],
	HUNTER = L["Hu"],
}

local unit -- this is set before the functions are called

local Modifiers; Modifiers = {
	-- when writing modifiers, inputs can be a number or a string. Also, any of these can be returned, as well as nil, which means a blank string.
	-- the modifier itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
	hidezero = function(value)
		if value ~= 0 then
			return value
		end
	end,
	
	negate = function(value)
		if type(value) == "number" then
			return -value
		end
		return value
	end,
	
	percent = function(value)
		if type(value) == "number" then
			return value .. "%"
		end
		return value
	end,
	
	hidemaxhp = { function(value)
		local maxValue = UnitHealthMax(unit)
		local currValue = UnitHealth(unit)
		if maxValue == 0 then
			return
		end
		local MobHealth3 = MobHealth3
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, 1, maxValue)
			if MHfound then
				if curr == max then
					return
				else
					return value
				end
			end
		end
		if currValue == maxValue then
			return
		else
			return value
		end
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	hidemaxmana = { function(value)
		if UnitMana(unit) == UnitManaMax(unit) then
			return
		end
		return value
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	hidenomana = { function(value)
		if UnitManaMax(unit) == 0 then
			return
		end
		return value
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	hidenocombat = { function(value)
		if UnitAffectingCombat(unit) then
			return value
		end
	end, "PLAYER_REGEN_ENABLED", "PLAYER_REGEN_DISABLED", "Update" },
	
	short = function(value)
		if type(value) ~= "number" then
			if value and value:match("/") then
				local a,b = value:match("^(%d+)/(%d+)$")
				a = Modifiers.short(tonumber(a))
				b = Modifiers.short(tonumber(b))
				if a and b then
					return a.."/"..b
				end
			end
			return value
		end
		if value >= 10000000 or value <= -10000000 then
			return ("%.1fm"):format(value / 1000000)
		elseif value >= 1000000 or value <= -1000000 then
			return ("%.2fm"):format(value / 1000000)
		elseif value >= 100000 or value <= -100000 then
			return ("%.0fk"):format(value / 1000)
		elseif value >= 10000 or value <= -10000 then
			return ("%.1fk"):format(value / 1000)
		else
			return value
		end
	end,
	
	veryshort = function(value)
		if type(value) ~= "number" then
			if value and value:match("/") then
				local a,b = value:match("^(%d+)/(%d+)$")
				a = Modifiers.veryshort(tonumber(a))
				b = Modifiers.veryshort(tonumber(b))
				if a and b then
					return a.."/"..b
				end
			end
			return value
		end
		if value >= 1000000 or value <= -1000000 then
			return ("%.0fm"):format(value / 1000000)
		elseif value >= 1000 or value <= -1000 then
			return ("%.0fk"):format(value / 1000)
		else
			return value
		end
	end,
	
	upper = function(value)
		if type(value) == "string" then
			return value:upper()
		end
		return value
	end,
	
	lower = function(value)
		if type(value) == "string" then
			return value:lower()
		end
		return value
	end,
	
	bracket = function(value)
		return ("[%s]"):format(value)
	end,
	
	paren = function(value)
		return ("(%s)"):format(value)
	end,
	
	angle = function(value)
		return ("<%s>"):format(value)
	end,
	
	brace = function(value)
		return ("{%s}"):format(value)
	end,
	
	hostilecolor = function(value)
		local r, g, b = GetHostilityColor(unit)
		return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
	end,
	
	classcolor = function(value)
		local r, g, b = GetClassColor(unit)
		return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
	end,
	
	hpcolor = { function(value)
		local r, g, b
		b = 0
		local perc = UnitHealth(unit) / UnitHealthMax(unit)
		if perc > 0.5 then
			g = 1
			r = 2 * (1 - perc)
		else
			g = perc * 2
			r = 1
		end
		return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	aggrocolor = function(value)
		local val = UnitReaction("player", unit) or 5
	
		local info = UnitReactionColor[val]
		return ("|cff%02x%02x%02x%s|r"):format(info.r * 255, info.g * 255, info.b * 255, value)
	end,
	
	difficultycolor = function(value)
		if UnitCanAttack("player", unit) then
			local level = UnitLevel(unit)
			if level == 0 then
				level = 99
			end
			local info = GetDifficultyColor(level)
			return ("|cff%02x%02x%02x%s|r"):format(info.r * 255, info.g * 255, info.b * 255, value)
		end
		return value
	end,
	
	manacolor = { function(value)
		local powerType = UnitPowerType(unit)
		local r,g,b
		if powerType == 0 then
			r,g,b = unpack(colors.mana)
		elseif powerType == 1 then
			r,g,b = unpack(colors.rage)
		elseif powerType == 2 then
			r,g,b = unpack(colors.focus)
		elseif powerType == 3 then
			r,g,b = unpack(colors.energy)
		else
			r,g,b = unpack(colors.unknown)
		end
		return ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
	end, "UNIT_DISPLAYPOWER" },
	
	combatcolor = { function(value)
		if UnitAffectingCombat(unit) then
			local r,g,b = unpack(colors.inCombat)
			return ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
		else
			return ("|r%s"):format(value)
		end
	end, "PLAYER_REGEN_ENABLED", "PLAYER_REGEN_DISABLED", "Update" },
	
	white = function(value)
		return ("|r%s"):format(value)
	end,
	red = function(value)
		return ("|cffff0000%s|r"):format(value)
	end,
	green = function(value)
		return ("|cff00ff00%s|r"):format(value)
	end,
	blue = function(value)
		return ("|cff0000ff%s|r"):format(value)
	end,
	cyan = function(value)
		return ("|cff00ffff%s|r"):format(value)
		end,
	fuchsia = function(value)
		return ("|cffff00ff%s|r"):format(value)
	end,
	yellow = function(value)
		return ("|cffffff00%s|r"):format(value)
	end,
	gray = function(value)
		return ("|cffafafaf%s|r"):format(value)
	end,
}

local Tags = {
	-- When writing tags, `unit` is modified before the call, that contains the unit id.
	-- Also, nil (which translates to empty string), a number, or a string can be returned.
	-- the tag itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
	name = { function()
		local name = UnitName(unit)
		if not name then
			return unit
		end
		return name
	end, "UNIT_NAME_UPDATE"},
	
	namehostile = "name:hostilecolor",
	
	status = {function()
		if UnitIsDead(unit) then
			return L["Dead"]
		elseif UnitIsGhost(unit) then
			return L["Ghost"]
		elseif not UnitIsConnected(unit) then
			return L["Offline"]
		elseif UnitAffectingCombat(unit) then
			return L["Combat"]
		elseif unit == "player" and IsResting() then
			return L["Resting"]
		end
	end, "Update" --[[due to the combat thing]] },
	
	statusgone = {function()
		if UnitIsDead(unit) then
			return L["Dead"]
		elseif UnitIsGhost(unit) then
			return L["Ghost"]
		elseif not UnitIsConnected(unit) then
			return L["Offline"]
		end
	end, "PLAYER_ALIVE", "PLAYER_DEAD", "PLAYER_UNGHOST" },
	
	statuscolor = {function()
		if UnitIsDeadOrGhost(unit) then
			local r,g,b = unpack(colors.dead)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		elseif not UnitIsConnected(unit) then
			local r,g,b = unpack(colors.disconnected)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		elseif UnitAffectingCombat(unit) then
			local r,g,b = unpack(colors.inCombat)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		elseif u == "player" and IsResting() then
			local r,g,b = unpack(colors.resting)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		else
			return "|r"
		end
	end, "Update" --[[due to the combat thing]]},
	
	happycolor = {function()
		local x = GetPetHappiness()
		if x == 3 then
			local r,g,b = unpack(colors.petHappy)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		elseif x == 2 then
			local r,g,b = unpack(colors.petNeutral)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		elseif x == 1 then
			local r,g,b = unpack(colors.petAngry)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		end
	end, "UNIT_HAPPINESS"},

	curhp = { function()
		local currValue = UnitHealth(unit)
		local MobHealth3 = MobHealth3
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local maxValue = UnitHealthMax(unit)
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
			if MHfound then
				return curr
			end
		end
		return currValue or 0
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	maxhp = { function()
		local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
		if maxValue == 0 then
			return 0
		end
		local MobHealth3 = MobHealth3
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue,maxValue)
			if MHfound then
				return max
			end
		end
		return maxValue
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	percenthp = { function()
		local hpmax = UnitHealthMax(unit)
		if hpmax == 0 then
			return 0
		end
		return ("%.0f"):format(UnitHealth(unit) / hpmax * 100)
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	missinghp = { function()
		local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
		if maxValue == 0 then
			return 0
		end
		local MobHealth3 = MobHealth3
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
			if MHfound then
				return max - curr
			end
		end
		return maxValue - currValue
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	fractionalhp = { function()
		local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
		local MobHealth3 = MobHealth3
		if maxValue > 100 then
			return currValue.."/"..maxValue
		elseif MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
			if MHfound then
				return curr.."/"..max
			end
		end
		return currValue.."/"..maxValue
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	smarthp = { function()
		local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
		local MobHealth3 = MobHealth3
		if maxValue > 100 then
			return currValue.."/"..maxValue
		end
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
			if MHfound then
				return curr.."/"..max
			end
		end
		return ("%.0f%%"):format(currValue / maxValue * 100)
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },

	curmana = { function()
		return UnitMana(unit) or 0
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	maxmana = { function()
		return UnitManaMax(unit)
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	percentmana = { function()
		local mpmax = UnitManaMax(unit)
		if mpmax == 0 then
			return 0
		end
		return ("%.0f"):format(UnitMana(unit) / mpmax * 100)
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	missingmana = { function()
		return UnitManaMax(unit) - UnitMana(unit)
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	fractionalmana = { function()
		return UnitMana(unit).."/"..UnitManaMax(unit)
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	typemana = { function()
		local p = UnitPowerType(unit)
		if p == 1 then
			return L["Rage"]
		elseif p == 2 then
			return L["Focus"]
		elseif p == 3 then
			return L["Energy"]
		else
			return L["Mana"]
		end
	end, "UNIT_DISPLAYPOWER" },
	
	level = { function()
		local x = UnitLevel(unit)
		if x > 0 then
			return x
		else
			return "??"
		end
	end, "UNIT_LEVEL" },
	
	class = function()
		return UnitClass(unit) or UNKNOWN
	end,
	
	smartclass = function()
		if UnitIsPlayer(unit) or (not UnitIsFriend("player", unit) and not UnitPlayerControlled(unit)) then
			return UnitClass(unit)
		end
	end,
	
	playerclass = function()
		if UnitIsPlayer(unit) then
			return UnitClass(unit)
		end
	end,
	
	shortclass = function()
		local _,c = UnitClass(unit)
		if c then
			return shortclasses[c]
		end
	end,
	
	shortsmartclass = function()
		if UnitIsPlayer(unit) or (not UnitIsFriend("player", unit) and not UnitPlayerControlled(unit)) then
			local _,c = UnitClass(unit)
			return shortclasses[c]
		end
	end,
	
	shortplayerclass = function()
		if UnitIsPlayer(unit) then
			local _,c = UnitClass(unit)
			return shortclasses[c]
		end
	end,
	
	classcolor = function()
		local r, g, b = GetClassColor(unit)
		return ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	end,
	
	creature = function()
		return UnitCreatureFamily(unit) or UnitCreatureType(unit) or UNKNOWN
	end,
	
	combos = { function()
		return GetComboPoints() or 0
	end, "PLAYER_COMBO_POINTS" },
	
	combosymbols = { function()
		return ("@"):rep(GetComboPoints())
	end, "PLAYER_COMBO_POINTS" },
	
	classification = function()
		if UnitClassification(unit) == "rare" then
			return L["Rare"]
		elseif UnitClassification(unit) == "rareelite" then
			return L["Rare-Elite"]
		elseif UnitClassification(unit) == "elite" then
			return L["Elite"]
		elseif UnitClassification(unit) == "worldboss" then
			return L["Boss"]
		end
	end,
	
	shortclassification = function()
		if UnitClassification(unit) == "rare" then
			return L["r"]
		elseif UnitClassification(unit) == "rareelite" then
			return L["r+"]
		elseif UnitClassification(unit) == "elite" then
			return L["+"]
		elseif UnitClassification(unit) == "worldboss" then
			return L["b"]
		end
	end,
	
	faction = function()
		local _, faction = UnitFactionGroup(unit)
		return faction
	end,
	
	connect = function()
		if not UnitIsConnected(unit) then
			return L["Offline"]
		end
	end,
	
	smartrace = function()
		if UnitIsPlayer(unit) then
			return UnitRace(unit) or UNKNOWN
		else
			return UnitCreatureFamily(unit) or UnitCreatureType(unit) or UNKNOWN
		end
	end,
	
	race = function()
		return UnitRace(unit)
	end,
	
	shortrace = function()
		local _,r = UnitRace(unit)
		return shortraces[r]
	end,
	
	pvp = function()
		return UnitIsPVP(unit) and L["PvP"]
	end,
	
	plus = function()
		return UnitIsPlusMob(unit) and "+"
	end,
	
	sex = function()
		local sex = UnitSex(unit)
		if sex == 0 then
			return L["Male"]
		elseif sex == 1 then
			return L["Female"]
		end
	end,
	
	shortsex = function()
		local sex = UnitSex(unit)
		if sex == 0 then
			return L["m"]
		elseif sex == 1 then
			return L["f"]
		end
	end,
	
	rested = {function()
		if GetRestState() then
			return L["Rested"]
		end
	end, "PLAYER_XP_UPDATE"},
	
	leader = function()
		if UnitIsPartyLeader(unit) then
			return L["L"]
		end
	end,
	
	leaderlong = function()
		if UnitIsPartyLeader(unit) then
			return L["Leader"]
		end
	end,
	
	happynum = { function()
		return GetPetHappiness() or 0
	end, "UNIT_HAPPINESS" },
	
	happytext = { function()
		return _G["PET_HAPPINESS" .. (GetPetHappiness() or 0)] or ""
	end, "UNIT_HAPPINESS" },
	
	happyicon = { function()
		local happy = GetPetHappiness()
		if happy == 3 then
			return ":D"
		elseif happy == 2 then
			return ":||"
		elseif happy == 1 then
			return "B("
		end
	end, "UNIT_HAPPINESS" },
	
	curxp = { function()
		return UnitXP(unit)
	end, "UNIT_PET_EXPERIENCE", "PLAYER_XP_UPDATE" },
	
	maxxp = { function()
		return UnitXPMax(unit)
	end, "UNIT_PET_EXPERIENCE", "PLAYER_XP_UPDATE" },
	
	percentxp = { function()
		local max = UnitXPMax(unit)
		if max > 0 then
			return ("%.0f"):format(UnitXP(unit) / max * 100)
		end
		return 0
	end, "UNIT_PET_EXPERIENCE", "PLAYER_XP_UPDATE" },
	
	missingxp = { function()
		return UnitXPMax(unit) - UnitXP(unit)
	end, "UNIT_PET_EXPERIENCE", "PLAYER_XP_UPDATE" },
	
	tappedbyme = { function()
		if UnitIsTappedByPlayer(unit) then
			return "*"
		end
	end, "Update" },

	istapped = { function()
		if UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit) then
			return L["Tapped"]
		end
	end, "Update" },
	
	fkey = function()
		local fkey
		if unit == "player" then
			fkey = 0
		else
			fkey = unit:match("^party(%d)$")
		end
		if fkey then
			return "F" .. (fkey + 1)
		end
	end,
	
	aggrocolor = function()
		local val = UnitReaction("player", unit) or 5
		
		local info = UnitReactionColor[val]
		return ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
	end,
	
	difficultycolor = function()
		if UnitCanAttack("player", unit) then
			local level = UnitLevel(unit)
			if level == 0 then
				level = 99
			end
			local info = GetDifficultyColor(level)
			return ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
		end
	end,
	
	manacolor = function()
		local powerType = UnitPowerType(unit)
		local r,g,b
		if powerType == 0 then
			r,g,b = unpack(colors.mana)
		elseif powerType == 1 then
			r,g,b = unpack(colors.rage)
		elseif powerType == 2 then
			r,g,b = unpack(colors.focus)
		elseif powerType == 3 then
			r,g,b = unpack(colors.energy)
		else
			r,g,b = unpack(colors.unknown)
		end
		return ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	end,
	
	combatcolor = { function()
		if UnitAffectingCombat(unit) then
			local r,g,b = unpack(colors.inCombat)
			return ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
		else
			return "|r"
		end
	end, "PLAYER_REGEN_ENABLED", "PLAYER_REGEN_DISABLED", "Update" },
	
	raidgroup = function()
		for i = 1, GetNumRaidMembers() do
			local name, rank, subgroup = GetRaidRosterInfo(i)
			if name == UnitName(unit) then
				return subgroup
			end
		end
	end,
	
	guild = function()
		return GetGuildInfo(unit)
	end,
	
	guildrank = function()
		local _, rank = GetGuildInfo(unit)
		return rank
	end,

	target = { function()
		local u = unit .. "target"
		if UnitExists(u) then
			return UnitName(u)
		end
	end, "UNIT_TARGET" },
	
	-- colors
	white = function()
		return "|r"
	end,
	red = function()
		return "|cffff0000"
	end,
	green = function()
		return "|cff00ff00"
	end,
	blue = function()
		return "|cff0000ff"
	end,
	cyan = function()
		return "|cff00ffff"
	end,
	fuchsia = function()
		return "|cffff00ff"
	end,
	yellow = function()
		return "|cffffff00"
	end,
	gray = function()
		return "|cffafafaf"
	end,
	
	smartmissinghp = "missinghp:hidezero:negate"
}

local runCode
do
	local function handler(literal, inner)
		local tag, modifiers = (":"):split(inner, 2)
		local func = Tags[tag]
		if func then
			local value
			if type(func) == "table" then
				value = func[1]()
			else
				value = func()
			end
			if modifiers then
				local mod, rest = nil, modifiers
				repeat
					mod, rest = (":"):split(rest, 2)
					if not value or value == "" then
						value = nil
					else
						local num = tonumber(value)
						if num then
							value = num
						end
					end
					local func = Modifiers[mod]
					if func then
						if value then
							if type(func) == "table" then
								value = func[1](value)
							else
								value = func(value)
							end
						end
					else
						func = Tags[mod]
						if func then
							if not value then
								if type(func) == "table" then
									value = func[1]()
								else
									value = func()
								end
							else
								return value
							end
						else
							return literal
						end
					end
				until not rest
			end
			return value or ""
		else
			return literal
		end
	end
	function runCode(code, u)
		unit = u
		return code:gsub("(%[([a-z_:]+)%])", handler)
	end
end
local codeToEventList = setmetatable({}, {__index = function(self, key)
	local t = new()
	for bit in key:gmatch("%[([a-z_:]+)%]") do
		local tag, modifiers = (":"):split(bit, 2)
		tag = Tags[tag]
		if tag then
			if type(tag) == "table" then
				for i = 2, #tag do
					local event = tag[i]
					if event then
						t[event] = true
					end
				end
			end
			if modifiers then
				for mod in modifiers:gmatch("[^:]+") do
					local modifier = Modifiers[mod] or Tags[mod]
					if type(modifier) == "table" then
						for i = 2, #tag do
							local event = modifier[i]
							if event then
								t[event] = true
							end
						end
					end
				end
			end
		end
	end
	self[key] = t
	return t
end, __mode = 'k'})

local antiAlias
do
	local function handler(literal, inner)
		local tag, modifiers = (":"):split(inner, 2)
		local func = Tags[tag]
		if func then
			local real = new()
			real[#real+1] = '['
			if type(func) == "string" then
				real[#real+1] = func
			else
				real[#real+1] = tag
			end
			if modifiers then
				for mod in modifiers:gmatch("[^:]+") do
					local func = Modifiers[mod] or Tags[mod]
					if func then
						real[#real+1] = ":"
						if type(func) == "string" then
							real[#real+1] = func
						else
							real[#real+1] = mod
						end
					else
						real = del(real)
						return literal
					end
				end
			end
			real[#real+1] = ']'
			local str = table.concat(real, '')
			real = del(real)
			return str
		else
			return literal
		end
	end
	antiAlias = setmetatable({}, {__index = function(self, key)
		local value = key:gsub("(%[([a-z_:]+)%])", handler)
		self[key] = value
		return value
	end, __mode= 'k'})
end

local eventData = {}

local registry
local codeHash

local function scheduleFunc()
	for text in pairs(eventData.Update) do
		if text:IsShown() then
			text:SetText(runCode(codeHash[text], registry[text]))
		end
	end
end
local function eventFunc(units, event)
	local checkUnits = next(units)
	for text in pairs(eventData[event]) do
		if text:IsShown() then
			local unit = registry[text]
			if not checkUnits or units[unit] then
				text:SetText(runCode(codeHash[text], unit))
			end
		end
	end
end
local NormalUnits = {player = true, pet = true, target = true, mouseover = true, focus = true}
for i = 1, 4 do
	NormalUnits['party' .. i] = true
	NormalUnits['partypet' .. i] = true
end
for i = 1, 40 do
	NormalUnits['raid' .. i] = true
end
function DogTag:RegisterFontString(text, unit, code)
	self:argCheck(text, 2, "table")
	self:argCheck(unit, 3, "string")
	self:argCheck(code, 4, "string")
	code = antiAlias[code]
	if codeHash[text] then
		if codeHash[text] == code and registry[text] == unit then
			if text:IsShown() then
				text:SetText(runCode(codeHash[text], registry[text]))
			end
			return
		end
		self:UnregisterFontString(text)
	end
	codeHash[text] = code
	registry[text] = unit
	if NormalUnits[unit] then
		local events = codeToEventList[code]
	
		for event in pairs(events) do
			local t = eventData[event]
			if not t then
				eventData[event] = new()
				t = eventData[event]
			
				if event == "Update" then
					self:ScheduleRepeatingEvent("DogTag-Updater", func, 0.15, scheduleFunc)
				else
					self:RegisterBucketEvent(event, 0.05, eventFunc, event)
				end
			end
			t[text] = true
		end
	end
	registry[text] = unit
	
	if text:IsShown() then
		text:SetText(runCode(code, unit))
	end
end

function DogTag:UnregisterFontString(text)
	self:argCheck(text, 2, "table")
	local code = codeHash[text]
	if not code then
		return
	end
	
	local unit = registry[text]
	registry[text] = nil
	codeHash[text] = nil
	if NormalUnits[unit] then
		local events = codeToEventList[code]
	
		for event in pairs(events) do
			local t = eventData[event]
			assert(t)
			t[text] = nil
			if not next(t) then
				eventData[event] = del(t)
				if event == "Update" then
					self:CancelScheduledEvent("DogTag-Updater")
				else
					self:UnregisterBucketEvent(event)
				end
			end
		end
	end
	
	text:SetText("")
end

function DogTag:UpdateFontString(text)
	self:argCheck(text, 2, "table")
	if registry[text] and text:IsShown() then
		text:SetText(runCode(codeHash[text], registry[text]))
	end
end

function DogTag:UpdateAllForUnit(unit)
	self:argCheck(unit, 2, "string")
	
	for text, u in pairs(registry) do
		if unit == u and text:IsShown() then
			text:SetText(runCode(codeHash[text], unit))
		end
	end
end

function DogTag:PARTY_MEMBERS_CHANGED()
	for text, unit in pairs(registry) do
		if text:IsShown() then
			text:SetText(runCode(codeHash[text], unit))
		end
	end
end

function DogTag:PLAYER_TARGET_CHANGED()
	self:UpdateAllForUnit('target')
end

function DogTag:PLAYER_FOCUS_CHANGED()
	self:UpdateAllForUnit('focus')
end

function DogTag:UPDATE_MOUSEOVER_UNIT()
	self:UpdateAllForUnit('mouseover')
end

function DogTag:PLAYER_PET_CHANGED()
	self:UpdateAllForUnit('pet')
end

local function activate(self, oldLib, oldDeactivate)
	DogTag = self
	
	AceEvent:embed(self)
	self:UnregisterAllEvents()
	self:UnregisterAllBucketEvents()
	self:RegisterEvent("PARTY_LEADER_CHANGED", "PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("RAID_ROSTER_UPDATE", "PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("PARTY_MEMBER_ENABLE", "PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("PARTY_MEMBER_DISABLE", "PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED")
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	self:RegisterEvent("PLAYER_PET_CHANGED")
	
	registry = {}
	self.registry = registry
	codeHash = {}
	self.codeHash = codeHash
	if oldLib then
		for text, unit in pairs(oldLib.registry) do
			self:RegisterFontString(text, unit, oldLib.codeHash[text])
		end
	end
end

AceLibrary:Register(DogTag, MAJOR_VERSION, MINOR_VERSION, activate, nil, external)
